open World
open Robots

let showd = false

let print_out s =
  if showd then (
    prerr_endline (">> " ^ s);
    flush stderr
  )

let print_in s =
  if showd then (
  prerr_endline ("<< " ^ s);
  flush stderr
  )

let debug s =
  if showd then (
    prerr_endline ("D: " ^ s)
  )

let print_package p = 
  debug (Packages.string_of_package p)

let print_packages ps =
  Packages.iter (fun p -> if Packages.PacSet.mem p (Packages.get_packages_by_location ps (Packages.Position p.Packages.destination)) then ()
else print_package p) ps

let print_packages_where ps =
  Packages.iter_where print_package ps

let print_packages_set ps =
  Packages.PacSet.iter print_package ps

let string_of_robot r =
  Printf.sprintf "robot: i %a; p %a; m %d; s %d"
    (fun () -> Robots.string_of_rid) r.ruid (fun () -> Utils.string_of_position) r.rposition r.money r.score

let print_robot p = 
  debug (string_of_robot p)
    
let print_robots robots =
  r_iter_on_robots robots (fun _ -> print_robot)

let string_of_worldcst wcst =
  Printf.sprintf "worldcst: s %a; m %d, c %d"
    (fun () -> Utils.string_of_position) (wcst.width, wcst.height) wcst.initial_money wcst.capacity

let print_world w =
  debug (Printf.sprintf "world: t %d" w.turn)

let print_worldcst wcst =
  debug (string_of_worldcst wcst)

let char_of_tile = function
  | World.Plain -> '.'
  | World.Lethal -> '~'
  | World.Impassable -> '#'
  | World.Home -> '@'

let print_map w =
  for i = w.cst.height + 1 downto 0 do
    for j = 0 to w.cst.width + 1 do
      prerr_char (char_of_tile w.cst.map.(i).(j))
    done;
    prerr_char '\n'
  done

let print_my_packages ps me =
  Packages.PacSet.iter print_package (Packages.packages_on_robot ps me)

let print_world w me = 
  print_world w;
  print_worldcst w.cst;
  debug "package_id:";
  print_packages w.packages;  
(*   debug "package_where:"; *)
(*   print_packages_where w.packages; *)
  debug "robots:";
  print_robots w.robots;
  debug "my_packages:";
  print_my_packages w.packages me
