type wherepackage =
  | Position of Utils.position
  | Robot of Robots.robot_id

type package = 
    { puid : int
    ; weight : int
    ; destination : Utils.position
    ; plocation : wherepackage
    }

module PacSet = Set.Make  (struct type t = package let compare = compare end)
module PacWhere = Map.Make (struct type t = wherepackage let compare = compare end)
module PacId = Map.Make (struct type t = int let compare = compare end)

type t = (package PacId.t * PacSet.t PacWhere.t)

let pacid_of_pac = fst
let pacwhere_of_pac = snd

let new_packages pacid pacwhere = (pacid, pacwhere)

let add_package_id ps p =
  PacId.add p.puid p (pacid_of_pac ps)

let add_package_where ps p =
  let pacwhere = pacwhere_of_pac ps in
  let set =
    try
      PacWhere.find p.plocation pacwhere 
    with Not_found ->
      PacSet.empty in
    PacWhere.add p.plocation (PacSet.add p set) pacwhere

let add_package ps p =
  new_packages 
    (add_package_id ps p)
    (add_package_where ps p)

let packages_of_list = 
  List.fold_left add_package (new_packages PacId.empty PacWhere.empty)

let add_packages p = 
  List.fold_left add_package p

let remove_package_id ps p =
  PacId.remove p.puid (pacid_of_pac ps)

let remove_package_where ps p =
  let pacwhere = pacwhere_of_pac ps in
    try
      PacWhere.add 
	p.plocation
	(PacSet.remove p (PacWhere.find p.plocation pacwhere))
	pacwhere
    with Not_found ->
      pacwhere

let rm_package ps p =
  new_packages
    (remove_package_id ps p)
    (remove_package_where ps p)

let get_package_by_id ps puid =
  try
    Some (PacId.find puid (pacid_of_pac ps))
  with Not_found -> None

let rm_package_by_id ps puid =
  match get_package_by_id ps puid with
      Some p -> rm_package ps p
    | None -> ps

let get_packages_by_location ps location =    
  try
    PacWhere.find location (pacwhere_of_pac ps)
  with Not_found -> PacSet.empty

let change_package_loc ps puid newloc =
  match get_package_by_id ps puid with
      Some p ->
	add_package (rm_package_by_id ps puid) { p with plocation = newloc }
    | None ->
	ps

let packages_on_robot ps r =
  try
    PacWhere.find (Robot r) (pacwhere_of_pac ps)
  with Not_found -> PacSet.empty

let fold_packages_on_robot ps r f =
  let robotloc = Robot r in
    PacSet.fold 
      (fun p acc -> f acc p)
      (packages_on_robot ps r)

let iter f packages =
  PacId.iter (fun _ p -> f p) (pacid_of_pac packages)

let iter_where f packages =
  PacWhere.iter (fun _ p -> (PacSet.iter f p)) (pacwhere_of_pac packages)

let list_of_packages ps = 
  PacId.fold (fun _ p acc -> p :: acc) (pacid_of_pac ps) []

let list_of_pacset ps = 
  PacSet.fold (fun p acc -> p :: acc)  ps []

let string_of_location = function
  | Position x -> "pos " ^ Utils.string_of_position x
  | Robot x -> "rob " ^ Robots.string_of_rid x

let string_of_package p =
  Printf.sprintf "package : i %d; p %a; w %d; l %a"
    p.puid (fun () -> Utils.string_of_position)  p.destination p.weight
    (fun () -> string_of_location) p.plocation
