type wherepackage = Position of Utils.position | Robot of Robots.robot_id
and package = {
  puid : int;
  weight : int;
  destination : Utils.position;
  plocation : wherepackage;
} 
module PacSet :
  sig
    type elt = package
    and t
    val empty : t
    val is_empty : t -> bool
    val mem : elt -> t -> bool
    val add : elt -> t -> t
    val singleton : elt -> t
    val remove : elt -> t -> t
    val union : t -> t -> t
    val inter : t -> t -> t
    val diff : t -> t -> t
    val compare : t -> t -> int
    val equal : t -> t -> bool
    val subset : t -> t -> bool
    val iter : (elt -> unit) -> t -> unit
    val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
    val for_all : (elt -> bool) -> t -> bool
    val exists : (elt -> bool) -> t -> bool
    val filter : (elt -> bool) -> t -> t
    val partition : (elt -> bool) -> t -> t * t
    val cardinal : t -> int
    val elements : t -> elt list
    val min_elt : t -> elt
    val max_elt : t -> elt
    val choose : t -> elt
  end
module PacWhere :
  sig
    type key = wherepackage
    and +'a t
    val empty : 'a t
    val add : key -> 'a -> 'a t -> 'a t
    val find : key -> 'a t -> 'a
    val remove : key -> 'a t -> 'a t
    val mem : key -> 'a t -> bool
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val map : ('a -> 'b) -> 'a t -> 'b t
    val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
    val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  end
module PacId :
  sig
    type key = int
    and +'a t
    val empty : 'a t
    val add : key -> 'a -> 'a t -> 'a t
    val find : key -> 'a t -> 'a
    val remove : key -> 'a t -> 'a t
    val mem : key -> 'a t -> bool
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val map : ('a -> 'b) -> 'a t -> 'b t
    val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
    val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  end

type t

val pacid_of_pac : t -> package PacId.t
val pacwhere_of_pac : t -> PacSet.t PacWhere.t

val add_package : t ->  PacSet.elt -> t

val packages_of_list :  PacSet.elt list -> t

val add_packages : t -> PacSet.elt list -> t

val rm_package : t -> package -> t

val get_package_by_id : t -> PacId.key -> package option

val rm_package_by_id : t ->  PacId.key -> t

val get_packages_by_location : t -> PacWhere.key -> PacSet.t

val change_package_loc : t -> PacId.key -> wherepackage -> t

val packages_on_robot : t -> Robots.robot_id -> PacSet.t

val fold_packages_on_robot :
  t -> Robots.robot_id -> ('a -> PacSet.elt -> 'a) -> 'a -> 'a

val iter : (package -> unit) -> t -> unit
val iter_where : (package -> unit) -> t -> unit

val list_of_packages : t -> package list

val list_of_pacset : PacSet.t -> package list

val string_of_location : wherepackage -> string

val string_of_package : package -> string
