module type OrderedType =
sig
    type t
    val compare: t -> t -> int
end

type priority = Inf | Int of int


let prio_compare x y =
  match  x, y with
      Int x, Int y -> compare x y
    | Int x, Inf -> -1
    | Inf, Int x -> 1
    | Inf, Inf -> 0

let lesser x y =
  prio_compare x y < 0

module Make (El : OrderedType) = struct

  let compare (x1, y1) (x2, y2) =
    match prio_compare y1 y2 with
      |	0 -> El.compare x1 x2
      | c -> c

  module PrioSet = Set.Make (struct type t = (El.t * priority) let compare = compare end)

  class t =
  object(self)
    val mutable prioset = PrioSet.empty
    val prio_hash : (El.t, priority) Hashtbl.t = Hashtbl.create 0

    method private set new_prioset =
      prioset <- new_prioset

    method priority el =
      Hashtbl.find prio_hash el
			 
    method add el prio =
      self#set (PrioSet.add (el, prio) prioset);
      Hashtbl.add prio_hash el prio
	
    method remove el =
      let prio = self#priority el in
	self#set (PrioSet.remove (el, prio) prioset);
	Hashtbl.remove prio_hash el

    method empty =
      prioset = PrioSet.empty
      
    method extract_min =
      let (el, _) as min = PrioSet.min_elt prioset in
	self#remove el;
	fst min
      
    method belong el =
      Hashtbl.mem prio_hash el
      
    method set_priority el prio =
      self#remove el;
      self#add el prio
  end
end
