type robot_id

type robot = {
  money : int;
  ruid : robot_id;
  rposition : Utils.position;
  score : int;
} 

module RobMap :
  sig
    type key = robot_id
    and +'a t
    val empty : 'a t
    val add : key -> 'a -> 'a t -> 'a t
    val find : key -> 'a t -> 'a
    val remove : key -> 'a t -> 'a t
    val mem : key -> 'a t -> bool
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val map : ('a -> 'b) -> 'a t -> 'b t
    val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
    val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  end
module IntSet :
  sig
    type elt = robot_id
    and t
    val empty : t
    val is_empty : t -> bool
    val mem : elt -> t -> bool
    val add : elt -> t -> t
    val singleton : elt -> t
    val remove : elt -> t -> t
    val union : t -> t -> t
    val inter : t -> t -> t
    val diff : t -> t -> t
    val compare : t -> t -> int
    val equal : t -> t -> bool
    val subset : t -> t -> bool
    val iter : (elt -> unit) -> t -> unit
    val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
    val for_all : (elt -> bool) -> t -> bool
    val exists : (elt -> bool) -> t -> bool
    val filter : (elt -> bool) -> t -> t
    val partition : (elt -> bool) -> t -> t * t
    val cardinal : t -> int
    val elements : t -> elt list
    val min_elt : t -> elt
    val max_elt : t -> elt
    val choose : t -> elt
  end

val r_empty : robot RobMap.t

val r_extract_me : robot RobMap.t -> RobMap.key -> robot * robot RobMap.t

val robot_pos : robot RobMap.t -> RobMap.key -> Utils.position

val r_add_robot : robot RobMap.t -> robot -> robot RobMap.t

val r_replace_robot : robot RobMap.t -> robot -> robot RobMap.t

val r_get_robot : robot RobMap.t -> RobMap.key -> robot

val r_rm_robot : robot RobMap.t -> RobMap.key -> robot RobMap.t

val r_set_pushed : IntSet.t -> IntSet.elt -> IntSet.t

val r_clear_pushed : IntSet.t

val r_was_pushed : IntSet.t -> IntSet.elt -> bool

val r_fold_on_robots :
  robot RobMap.t -> (RobMap.key -> robot -> 'b -> 'b) -> 'b -> 'b

val r_iter_on_robots : robot RobMap.t -> (RobMap.key -> robot -> unit) -> unit

val string_of_rid : robot_id -> string
val unsafe_rid_of_int : int -> robot_id
