open Strategie

module G = Graphs.Make (struct type t = int * int let compare = compare end)

let adj_of_pos world pos =
  let size = Array.length pos in
  let adj = Array.create_matrix size size 0 in
    for i = 0 to size - 1 do
      let comp = get_path_computer world pos.(i) in
	for j = 0 to i - 1 do
	  let dist = get_dist comp pos.(j) in
	    adj.(i).(j) <- dist;
	    adj.(j).(i) <- dist;
	done
    done;
    adj

let uniqify l = 
  let h = Hashtbl.create 0 in
  List.iter (fun e -> Hashtbl.replace h e ()) l;
  Hashtbl.fold (fun k () acc -> k :: acc) h []

let salesman world start positions n =
  if positions = [] then
    (0, [])
  else begin
    let h = Hashtbl.create 0 in
    let _ = List.iter (fun (pos, ob) -> Hashtbl.add h pos ob) (positions) in
    let pos_ar = Array.of_list (uniqify (start::List.map fst positions)) in
    let adj = adj_of_pos world pos_ar in
    let g = G.create pos_ar adj in
    let path = (if G.size g <= n then ((* Debug.debug "salesman exhaustif !!";  *)G.salesman_ex) else G.salesman) g in
    let cost = G.cost g path in
      try
	   (cost, List.map (fun p -> (p, Hashtbl.find h p)) (List.tl
							       (Array.to_list path)))
      with Not_found ->
	(G.cost g pos_ar, positions)
  end
	   
