open World
open Robots
open Debug    
open Packages
open Printf
open Strategie
open Utils

class virtual sta = Strategie.stratacc

class ['a, 'b] meta_then (s1:'a) (s2:'b) = object (self)
  inherit sta

  method id = "meta_then" 

  val s = s1 
  method nextmove_core w =
    match s#nextmove w with
      _, None -> debug ("meta_then : switch to " ^ s2#id);  s2#nextmove w
    | s', ac -> ({< s = s' >} :> sta), ac
end

class ['a, 'b] meta_sub (prio:'a) (low:'b) = object
  inherit sta

  method id = "meta_sub"

  val sprio = prio
  val slow = low
  method nextmove_core w =
    match sprio#nextmove w with
      newsprio, None -> 
	debug ("meta_sub : switch to " ^ slow#id);
	let newslow, cmd = slow#nextmove w in
	({< sprio = newsprio; slow = newslow >} :> sta), cmd
    | newsprio, cmd -> ({< sprio = newsprio >} :> sta), cmd
end

class ['a, 'b] meta_cycle (sinit:'a) = object
  val s = sinit
  method nextmove world =
    match s#nextmove world with
      Some nexts, ac -> Some {< s = nexts >}, ac
    | None, ac -> Some {< s = sinit >}, ac
end

let meta_then a b = new meta_then a b
let meta_sub a b = new meta_sub a b

class nothing_strat = object (self)
  inherit sta 
  method id = "nothing_strat"
  method nextmove_core world =
    (self:> sta), None
end

class goto_strat_helper world pos me strat_to_apply =
  let my_pos = (get_robot world me).rposition in
  let path_computer = Strategie.get_path_computer world.cst my_pos in
  let path = 
    try
      Some (my_pos :: Strategie.get_path path_computer pos)
    with Not_found -> 
      None in
object(self)
  inherit sta

  val pos = pos
  val path = path
  val me = me
  val world = world
		
  method my_pos = (get_robot world me).rposition

  method id = "goto_strat_helper"

  method nextmove_core world =
    let new_goto () =
      (new goto_strat_helper world pos me strat_to_apply :>
       Strategie.stratacc)#nextmove world in
    debug (sprintf "going to %s" (string_of_position pos));
    if self#my_pos = pos then
      (strat_to_apply  :> Strategie.stratacc)#nextmove world
    else
      match path with
      | None -> ((self : #Strategie.stratacc :> Strategie.stratacc), None)
      | Some (hd1 :: hd2 :: tl) ->
	  if self#my_pos = hd1 then
	    (({< path = Some (hd2 :: tl) >} :> Strategie.stratacc), Some ((Move (get_move self#my_pos hd2)), 1))
	  else
	    new_goto ()
      | _ ->
	  new_goto ()
end 

class goto_strat pos me strat_to_apply =
object(self)
  inherit sta

  method id = "goto_strat"

  method nextmove_core world =
    (new goto_strat_helper world pos me strat_to_apply)#nextmove world
end


class drop me = object (self)
  inherit sta

  method id = "drop"

  method nextmove_core world =
    let my = get_robot world me in
    let mypkg = packages_on_robot world.packages me in
    let lpacks = list_of_pacset mypkg in
    let l = List.filter (fun p -> p.destination = my.rposition) lpacks in
    if l = [] then (self :> sta), None
    else (self :> sta), Some (Drop (List.map (fun p -> p.puid) l), 1)
    
end


module BaseMap = Map.Make (struct type t = Utils.position let compare = compare end)

class at_home_bases_helper world me (strat : sta) last_base =
  let bases_pos = all_homebases world in
  let _ = List.iter (fun bp -> debug ("home base : " ^ string_of_position bp)) bases_pos in
object(self)
  inherit sta

  method id = "at_home_bases_helper"

  val visited = []

  method closest_base world =    
    let bases = List.filter 
		  (fun bp -> 
		     not (List.mem bp visited) ||
		     get_packages_by_location world.packages (Position bp) <> PacSet.empty
		  ) bases_pos in
    let posclosest = Strategie.get_closest world.cst (get_robot world me).rposition bases in
    posclosest

  method nextmove_core world =
    let my = get_robot world me in
      match self#closest_base world with
	Some closest ->
	  if closest = my.rposition then
	    (meta_then strat {< visited = closest :: visited >})#nextmove world
	  else
	    (new goto_strat closest me self)#nextmove world
      | None -> 
	  if List.mem my.rposition bases_pos then 
	  (meta_then strat (new nothing_strat))#nextmove world
	  else
	    (new nothing_strat :> sta), None
(* 	  if List.mem my.rposition bases_pos then *)
(* 	    (meta_then strat {< visited = my.rposition :: visited >})#nextmove world *)
(* 	  else *)
(* 	    (self :> sta), None *)
    
end


let pick_glouton world me wmax pkg_delivrables = 
  let l = list_of_pacset pkg_delivrables in
  Strategie.pick_packages wmax l
    
    
class pick algo me = object (self)
  inherit sta

  method id = "pick"

  method nextmove_core world =
    let my = get_robot world me in
    let pkgs = get_packages_by_location world.packages (Position my.rposition) in
    let paths = Strategie.paths world.cst my.rposition in
(*     let pkgs = PacSet.filter (fun p -> p.destination <> my.rposition) pkgs in *)
(*     let pkgs_deliv = PacSet.filter (fun p -> Hashtbl.mem paths p.destination) pkgs in *)
(*     debug "Picking among packages :"; *)
(*     print_packages_set pkgs; *)
    let myweight = weight_of_robot world me in
    let lpick = algo world me (world.cst.capacity - myweight) pkgs in
    if lpick = [] then (
      debug "nothing to pick";
      (self :> sta), None
    )
    else       
      (self :> sta), Some (Pick (List.map (fun x -> x.puid) lpick), 1)
end   

let at_home_bases world me strat =
  new at_home_bases_helper world me strat None


let iter_by_subset maxstep f my_compare world me wmax pkgs = 
  let l = list_of_pacset pkgs in
  let l = List.fast_sort my_compare l in
  let nbstep = ref 0 in
  let wmaxsofar = ref 0 in
  let rec pick picked weightsofar l = 
    incr nbstep;
    if !nbstep > maxstep then ()
    else match l with
      [] -> if weightsofar >= !wmaxsofar then (
	f picked weightsofar;
	wmaxsofar := weightsofar
      )
    | x::xs -> 	
	if x.weight + weightsofar > wmax then pick picked weightsofar xs
	else (
	  pick (x::picked) (x.weight + weightsofar) xs;	  	  
	  pick picked weightsofar xs
	)
  in
    pick [] 0 l

let iter_by_subset_weight_increasing maxstep f =
  iter_by_subset maxstep f (fun p1 p2 -> compare p2.weight p1.weight)

let iter_by_location_weight maxstep f =
  iter_by_subset maxstep f (fun p1 p2 ->
		      match compare p1.destination p2.destination with
			  0 -> compare p2.weight p1.weight
			| n -> n)

exception Break

let pick_maximizing ?(iter=iter_by_subset_weight_increasing)
  ftransf fbetter maxiter world me wmax pkg_delivrables =
  let best = ref None in
	 iter maxiter
	   (fun p wsf ->
	      match !best with
		   None -> best := Some (ftransf p wsf)
		 | Some old -> let newone = ftransf p wsf in
		     if fbetter newone old then best := Some newone
	      
	   )
	   world me wmax pkg_delivrables;
    match !best with
	None -> ftransf [] 0 (* should not happen, iter calls its function at least once *)
      | Some x -> x
	  
let pick_unpeumieux world me wmax pkg_delivrables = 
  fst (pick_maximizing (fun x y -> x, y) (fun (_, wnew) (_, wold) -> wnew > wold) 10000 world me wmax pkg_delivrables)



class try_goto world me pos strat_error strat_ok =
  let my_pos = (get_robot world me).rposition in
  let path_computer = Strategie.get_path_computer world.cst my_pos in
  let path = 
    try
      Some (my_pos :: Strategie.get_path path_computer pos)
    with Not_found -> 
      None in
  let _ = 
    debug (sprintf "trying to go to %s" (string_of_position pos))
  in
object
  inherit sta

  method id = "try_goto"

  val path = path
  method nextmove_core world =
    let my_pos = (get_robot world me).rposition in
    if my_pos = pos then strat_ok#nextmove world
    else
      match path with
      | Some (p1::p2::xs) when p1 = my_pos -> 
	  ({< path = Some (p2::xs) >} :> sta), Some ((Move (get_move my_pos p2)), 1)
      | _ -> strat_error#nextmove world
	  
end

class try_deliver me (l, ps) strat_error = object (self)
  inherit sta

  method id = "try_deliver"

  val l = l
  val ps = ps
  method nextmove_core world =
    
    match l with
      [] -> (self :> sta), None
    | p::xs -> 
	if PacSet.equal (packages_on_robot world.packages me) ps then
	  (* si p est encore dans la liste *)
	  (new try_goto world me p.destination self 
	     (meta_then (new drop me) 
		(let notinpdest = List.filter (fun x -> x.destination <> p.destination) xs in
		 let newps = PacSet.fold 
			       (fun x acc -> 
				  if p.destination = x.destination 
				  then PacSet.remove x acc
				  else acc) ps ps in
		 {< l = notinpdest; ps = newps >})
	   :> sta
	  ))#nextmove world
	else
	  strat_error#nextmove world
end


class deliver_all delivery_order_packages me = object (self)
  inherit sta

  method id = "deliver_all"

  method nextmove_core world =
    let mypkg = packages_on_robot world.packages me in
    if mypkg = PacSet.empty then (* pas de pkg  dlivrer *)
      (self :> sta), None
    else
      let l = delivery_order_packages world me mypkg in
      (new try_deliver me (l, mypkg) self)#nextmove world
end

class deliver_stratbase me = object (self)
  inherit sta

  method id = "deliver_stratbase"

  method nextmove_core world =
    let mypkg = packages_on_robot world.packages me in
    if mypkg = PacSet.empty then (* pas de pkg  dlivrer *)
      (self :> sta), None
    else
      let my = get_robot world me in
      match closest world.cst my.rposition (List.map (fun x -> x.destination) (list_of_pacset mypkg)) with
	Some (_, revpath) ->
	  (match revpath with
	     [] -> (new drop me)#nextmove world
	   | x::_ -> (new try_goto world me x self (meta_then (new drop me) self))#nextmove world
	  )
      | None -> 
	  (self :> sta), None
end

let stratbase_old world me =
  at_home_bases world me (meta_then (new pick pick_glouton me) 
			    (new deliver_stratbase me :> sta)
			 )

let super_algo_de_livraison_avec_cout world me ps =
  let my = get_robot world me in
  let l = PacSet.fold (fun p l -> (p.destination, p) :: l) ps [] in
(*   debug "appel du salesman"; *)
(*   debug ("ma pos :" ^ string_of_position my.rposition); *)
(*   List.iter (fun (dest, p) -> print_package p) l; *)
  let cost, tour = Salesman.salesman world.cst my.rposition l 8 in
(*   debug ("cout du salesman:" ^ string_of_int cost); *)
(*   if tour = [] then debug "hum... liste vide"; *)
  let tour = List.map snd tour in
  (* calcule la version idiote *)
  let rec closest_first sumcost pos l =
    match closest world.cst pos l with
      Some (cost, closest) -> 
	(match closest with
	   x::_ ->
	     let c, l = closest_first (cost + sumcost) x (List.filter ((<>) x) l) in
	     c, x :: l
	 | [] -> 1000000, []
	)
    | None -> sumcost, []
  in
  let h = Hashtbl.create 0 in
  List.iter (fun (pos, pac) -> Hashtbl.add h pos pac) l;
  let costdumb, tourdumb = closest_first 0 my.rposition (List.map fst l) in
(*   debug ("cout du dumb:" ^ string_of_int costdumb); *)
  if costdumb < cost || cost = 0 then
    costdumb, List.map ((fun pos -> Hashtbl.find h pos):position->package) tourdumb
  else 
    cost, tour

let super_algo_de_livraison world me ps =
  snd (super_algo_de_livraison_avec_cout world me ps)

class pick_when_can me picker = 
object (self)
  inherit sta

  method id = "pick_when_can"

  method nextmove_core world =
    let my = get_robot world me in
    if get_packages_by_location world.packages (Position my.rposition) <> PacSet.empty then (
      debug "trying to pick";
      let _, ac = (new pick picker me)#nextmove world in
      (self :> sta), ac
    )
    else (self :> sta), None
end

class mega_pick me =
object(self)
  inherit sta

  method id = "mega_pick"
		
  method nextmove_core world =
    let my = get_robot world me in
    if packages_on_robot world.packages me = PacSet.empty 
      && get_packages_by_location world.packages (Position my.rposition) <> PacSet.empty
    then 
      begin
	let ftransf packages weight =
	  let pack = List.fold_right (fun x acc -> PacSet.add x acc)
		       packages PacSet.empty in
	  let (cost, _) = super_algo_de_livraison_avec_cout world me pack in
	  (* paranoid *) 
	  if cost = 0 then (max_float, pack)
	  else
	    (float_of_int cost /. float_of_int (weight + 1), pack) in
	(* 	    (float_of_int cost /. float_of_int weight, pack) in *)
	let maxiter = 1000 in
	let fbetter (x, _) (y, _) = x < y in
	let get_better x y = if fbetter x y then x else y in
	let algo world me capacity pkgs_deliv =
	  let s1 = pick_maximizing ftransf fbetter maxiter world me 
 		     capacity pkgs_deliv in 
	  let s2 =  pick_maximizing ~iter:iter_by_location_weight 
		      ftransf fbetter maxiter world me capacity pkgs_deliv in
	  PacSet.fold (fun x acc -> x :: acc) (snd (get_better s1 s2)) [] in
	let _, ac = (new pick algo me)#nextmove world in
	(self :> sta), ac
      end 
    else
      ((self :> sta), None)
end

class ultra_pick me =
object(self)
  inherit sta

  method id = "ultra_pick"

  method nextmove_core world =
    let my = get_robot world me in
    let my_pack = packages_on_robot world.packages me in
      if get_packages_by_location world.packages (Position my.rposition) <> PacSet.empty
      then 
	begin
	  if get_packages_by_location world.packages (Robot me) <> PacSet.empty then
	    debug "ultrapicking !!!";
	  let ftransf packages weight =
	    let pack = List.fold_right (fun x acc -> PacSet.add x acc)
			 packages PacSet.empty in
	    let (cost, _) = super_algo_de_livraison_avec_cout world me (PacSet.union pack my_pack) in
	      (* paranoid *) 
	      if cost = 0 then (max_float, pack)
	      else
		(float_of_int cost /. float_of_int (weight + 1), pack) in
	    (* 	    (float_of_int cost /. float_of_int weight, pack) in *)
	  let maxiter = 1000 in
	  let fbetter (x, _) (y, _) = x < y in
	  let get_better x y = if fbetter x y then x else y in
	  let algo world me capacity pkgs_deliv =
	    let s1 = pick_maximizing ftransf fbetter maxiter world me 
 		       capacity pkgs_deliv in 
	    let s2 =  pick_maximizing ~iter:iter_by_location_weight 
			ftransf fbetter maxiter world me capacity pkgs_deliv in
	      PacSet.fold (fun x acc -> x :: acc) (snd (get_better s1 s2)) [] in
	  let _, ac = (new pick algo me)#nextmove world in
	    (self :> sta), ac
	end 
      else
	((self :> sta), None)
end

let stratbase world me =
  meta_sub (new pick_when_can me pick_glouton)
    (at_home_bases world me (new deliver_all super_algo_de_livraison me :> sta) :> sta)
    

let notbad1 world me =
(*   meta_sub (new pick_when_can me pick_glouton) *)
  meta_sub (new mega_pick me)
    (at_home_bases world me (new deliver_all super_algo_de_livraison me :> sta) :> sta)
    
class basic world me = 
object (self)
    inherit sta
      
    method id = "basic"

    val bases = all_homebases world 
		  
    method nextmove_core world =
      (* si on a des packages, on les trie et on les dlivre *)
      let my = get_robot world me in
      let mypkg = packages_on_robot world.packages me in
      let donothing = (self :> sta), None in
    PacSet.iter (fun p -> debug (string_of_package p)) mypkg;
	if mypkg = PacSet.empty then (
	  (* pas de pkg -> goto base *)      
	  match Strategie.closest world.cst my.rposition bases with
	    | Some (_, revpath) -> 
		(* il y a une base proche *)
		(match List.rev revpath with
		     [] -> (* on est sur une base, on pick ce qu'on peut *)	       
	       let pkgs = get_packages_by_location world.packages (Position my.rposition) in
	       let paths = Strategie.paths world.cst my.rposition in
	       let pkgs = PacSet.filter (fun p -> p.destination <> my.rposition) pkgs in
	       let pkgs_deliv = PacSet.filter (fun p -> Hashtbl.mem paths p.destination) pkgs in
	       let myweight = weight_of_robot world me in
	       let lpick = pick_unpeumieux world me (world.cst.capacity - myweight) pkgs_deliv in
	       if lpick = [] then
		 ({< bases = List.filter (fun b -> b <> my.rposition) bases >}  :> sta)#nextmove world
	       else
		 (self :> sta), Some (Pick (List.map (fun x -> x.puid) lpick), 1)
	   | xy::_ -> (* on va vers la base *)
	       (self :> sta), Some (Move (get_move my.rposition xy), 1)
	  )
      | None -> (* base inaccessible *)
	  donothing
    ) else (
      (* on doit dlivrer les packages *)
      let lpacks = list_of_pacset mypkg in
      (* on cherche le plus proche *)
      (match Strategie.closest world.cst my.rposition (List.rev_map (fun p -> p.destination) lpacks) with
       | Some (_, revpath) -> 
	   (* le plus proche est par l *)
	   (match List.rev revpath with
	      [] -> (* le(s) plus proche est(sont) en dessous *)
		let l = List.filter (fun p -> p.destination = my.rposition) lpacks in
		(self :> sta), Some (Drop (List.map (fun p -> p.puid) l), 1)
	      | xy::_ -> (* on va vers le plus proche *)
		(self :> sta), Some (Move (get_move my.rposition xy), 1)
	   )
       | None -> (* aucun n'est dlivrable *)
	   (self :> sta), Some (Drop (List.map (fun p -> p.puid) lpacks), 1)
      )
    )
end

exception Sigalarm
let after n f g =
  try
    Sys.set_signal Sys.sigalrm (Sys.Signal_handle (fun _ -> raise Sigalarm));
    ignore (Unix.alarm n);
    let x = f () in
      Sys.set_signal Sys.sigalrm (Sys.Signal_ignore);
      ignore (Unix.alarm 0);
      x
  with Sigalarm -> g ()

class ['a,'b] dontfoire maxfoire (stbase:sta) (stbackup:sta) = object (self)
  inherit sta

  method id = "dontfoire"

  val nxtbase = stbase
  val nxtbackup = stbackup
  val nbbasefoired = 0
  method nextmove_core world =
    let nbs = (world.turn + 10) - (int_of_float (Sys.time ())) in
    (* to be cool *)
    let nbs = min 60 nbs in
    if nbbasefoired < maxfoire && nbs > 0 then
      after nbs
	(fun () -> let x, ac = nxtbase#nextmove world in ({< nxtbase = x >} :> sta), ac)
	(fun () -> 
	   let x, ac = nxtbackup#nextmove world in ({< nxtbackup = x; nbbasefoired = nbbasefoired + 1 >} :> sta), ac)
    else (
      debug "*********\n  WARNING :::::::::::: => \n*************\n BACKUP STRATEGY !!!!!!!!!!\n************";
      stbackup#nextmove world
    )
end

class push me ~(bid: int -> int) ~max_player =
object(self)
  inherit sta

  method id = "push"

  method nextmove_core world =
    let do_nothing = ((self :> sta), None) in
    let nb_robot = number_of_robot world in
      if nb_robot > max_player then
	do_nothing
      else begin
	let (x, y) as my_pos = robot_pos world.robots me in
	let can_be_pushed d = 
	  let pos = next_pos d my_pos in
	  match robot_of_pos world pos with
	  | Some r -> 
	      let x, y = next_pos d pos in
	      world.cst.map.(y).(x) = Lethal
	  | None -> false in
	match List.filter can_be_pushed ([N; S; W; E]) with
	| h :: t -> ((self :> sta), Some (Move h,
					  (bid (money_of_robot world me))))
	| [] -> do_nothing
      end
end 

class ['a] vice me (strat : 'a)= object
  inherit sta

  method id = "vice"

  val strat = strat

  method nextmove_core world =
    let ortho = function
	N | S -> [W; E]
      | W | E -> [N; S] in

    match strat#nextmove world with
      | (s, ((Some (Move d, _)) as ac)) as ret ->
	  (try
	     let pos = robot_pos world.robots me in
	     let pos1 = next_pos d pos in  
	     let (x, y) as pos2 = next_pos d pos1 in
	     let do_vice =
	       if world.cst.map.(y).(x) = Lethal then
		 List.exists (fun d ->
				match robot_of_pos
				  world (next_pos d pos1) with
				    | Some r -> true
				    | None -> false) (ortho d)
	       else
		 false in
	       if do_vice then 
		 ({< strat = s >} :> sta), Some (Move d, -2)
	       else
		 ({< strat = s >} :> sta), ac
	   with _ -> ({< strat = s >} :> sta), ac
	  )
      | (s, ac) -> ({< strat = s >} :> sta), ac
end


class explorer me = object (self)
  inherit sta

  method id = "explorer"

  method nextmove_core world =
    let x = Random.int world.cst.width + 1 in
    let y = Random.int world.cst.height + 1 in
    (new try_goto world me (x, y) self self)#nextmove world
end

let ultimatestrat w m =
  meta_sub 
    (new push m ~bid:(fun _ -> 2) ~max_player:4)
    (meta_sub 
       (new vice m (notbad1 w m :> sta))
       (new explorer m)
    )  

let resistant w m =
  new dontfoire 5 (ultimatestrat w m) (stratbase w m :> sta)

class dum m = object
  inherit sta
  method id = "dum"
  method nextmove_core w =
    (new goto_strat (4,11) m (new drop m) :> sta), 
    Some (Pick [0], 1)
end

class die = object (self)
  inherit sta
  method id = "die"
  method nextmove_core w =
    (self :> sta), Some (Pick [], 100000)
end

let checkscore w m = 
  meta_then (new goto_strat (11, 11) m 
	       (new dum m))
    (new die)

class faineasse m = object (self)
  inherit sta
  method id = "faineasse"
  method nextmove_core world =
    let my = get_robot world m in
    if my.rposition = (1, 8) then
      (self :> sta), None
    else
      (new goto_strat (1,8) m self)#nextmove world
end

